local AddonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local c = BittensSpellFlashLibrary

local GetPowerRegen = GetPowerRegen
local IsMounted = IsMounted
local UnitExists = UnitExists
local math = math

c.Init(a)

c.AssociateTravelTimes(
	.75,
	"Aimed Shot", 
	"Arcane Shot", 
	"Chimera Shot", 
	"Cobra Shot",
	"Serpent Sting",
	"Steady Shot") 

local function sufficientResources(z)
	return c.GetCost(z.ID) <= a.Focus
end

local function modSpell(spell)
	spell.EvenIfNotUsable = true
	spell.NoPowerCheck = true
	spell.CheckLast = sufficientResources
end

local function addSpell(name, tag, attributes)
	modSpell(c.AddSpell(name, tag, attributes))
end

local function addOptionalSpell(name, tag, attributes)
	modSpell(c.AddOptionalSpell(name, tag, attributes))
end

------------------------------------------------------------------------ Common
local markDebuffs = c.GetIDs("Hunter's Mark", "Marked for Death")

c.AddSpell("Mend Pet", nil, {
	FlashColor = "red",
	CheckFirst = function()
		return s.HealthPercent("pet") < 80
			and not s.Buff(c.GetID("Mend Pet"), "pet", c.GetBusyTime() + 2)
	end
})

c.AddOptionalSpell("Hunter's Mark", nil, {
	Debuff = markDebuffs,
	FlashColor = "yellow",
})

addOptionalSpell("Serpent Sting", nil, {
	CheckFirst = function(z)
		return c.ShouldCastToRefresh(
			"Serpent Sting", "Serpent Sting", 0, true, "Cobra Shot")
	end
})

addSpell("Arcane Shot")

c.AddSpell("Arcane Shot", "over 64", {
	CheckFirst = function()
		return a.Focus / s.MaxPower("player") > .64
			or c.HasBuff("The Beast Within")
	end
})

addSpell("Multi-Shot")

addOptionalSpell("Explosive Trap", nil, {
	singleID = "Explosive Trap",
	doubleID = { "Explosive Trap", "Trap Launcher" },
	Run = function(z)
		if c.GetCost("Trap Launcher") <= a.Focus then
			z.FlashID = z.doubleID
		else
			z.FlashID = z.singleID
		end
	end
})

c.AddOptionalSpell("Aspect of the Hawk", nil, {
	Type = "form",
	NotWhileMoving = 1,
	CheckFirst = function()
		return not s.Form("Aspect of the Wild")
	end
})

c.AddOptionalSpell("Kiroptyric Sigil", nil, {
	Type = "item",
	NoGCD = 1,
})

c.AddOptionalSpell("Call Pet", nil, {
	ID = "Call Pet 1",
	FlashID = {
		"Call Pet 1",
		"Call Pet 2",
		"Call Pet 3",
		"Call Pet 4",
		"Call Pet 5",
		"Revive Pet",
	},
	CheckFirst = function()
		return not IsMounted() and not UnitExists("pet")
	end
})

----------------------------------------------------------------- Beast Mastery
c.AddOptionalSpell("Focus Fire", nil, {
	CheckFirst = function()
		return not c.HasBuff("Focus Fire")
			and s.BuffStack(c.GetID("Frenzy Effect"), "pet") == 5
			and not c.HasBuff("The Beast Within")
			and c.GetCooldown("Bestial Wrath") > 10
	end
})

addSpell("Kill Command")

c.AddOptionalSpell("Bestial Wrath", nil, {
	CheckFirst = function()
		return a.Focus > 70
			and s.BuffStack(c.GetID("Frenzy Effect"), "pet") == 5
	end
})

c.AddOptionalSpell("Bestial Wrath", "for AoE", {
	CheckFirst = function()
		local _, regen = GetPowerRegen()
		return a.Focus > s.MaxPower("player") - a.FocusAdded() - regen
	end
})

addSpell("Multi-Shot", "for BM", {
	CheckFirst = function()
		if c.GetCooldown("Bestial Wrath") > 2 then
			return true
		end
		
		local _, regen = GetPowerRegen()
		return a.Focus > s.MaxPower("player") - a.FocusAdded() - regen
	end
})

c.AddOptionalSpell("Fervor", nil, {
	CheckFirst = function()
		return a.Focus < 50 and s.Power("pet") < 50
	end
})

c.AddSpell("Burrow Attack", nil, {
	Type = "pet",
	Continue = true,
})

c.AddSpell("Froststorm Breath", nil, {
	Type = "pet",
	Continue = true,
})

------------------------------------------------------------------ Marksmanship
local function underCarefulAim()
	return s.HealthPercent() > 90
		and c.HasTalent("Careful Aim")
		and not c.HasBuff("Aspect of the Fox")
		and not a.GetConfig("no_careful_aim")
end

addOptionalSpell("Serpent Sting", "unless Careful Aim", {
	CheckFirst = function(z)
		if underCarefulAim() then
			return false
		end
		
		local earlyRefresh = 0
		if c.GetCooldown("Chimera Shot")
			> c.GetMyDebuffDuration("Serpent Sting")
				+ c.EstimateTravelTime("Chimera Shot") then
			
			earlyRefresh = 3
		end
		return c.ShouldCastToRefresh(
			"Serpent Sting", "Serpent Sting", earlyRefresh, true,
			"Chimera Shot")
	end
})

c.AddSpell("Chimera Shot", nil, { 
	Override = function(z)
		if a.CSCool > .1 or a.Focus < c.GetCost("Chimera Shot") then
			return false
		end
		
		if not underCarefulAim() then
			return true
		end
		
		return c.HasTalent("Marked for Death")
			and s.DebuffDuration(markDebuffs)
				< c.GetBusyTime() + c.EstimateTravelTime("Chimera Shot")
	end
})

addSpell("Aimed Shot")

c.AddSpell("Aimed Shot!", nil, {
	FlashID = { "Aimed Shot", "Aimed Shot!" },
	Override = function()
		return c.HasBuff("Fire!")
	end
})

c.AddSpell("Marksmanship Focus Dump", nil, {
	ID = 0,
	Override = function(z)
		z.ID = c.GetID("Arcane Shot")
		if not c.HasBuff("Aspect of the Fox") then
			local aimedID = c.GetID("Aimed Shot")
			if underCarefulAim() then
				z.ID = aimedID
				return a.Focus >= c.GetCost("Aimed Shot")
			end
			
			local cast = c.GetCastTime("Aimed Shot")
			if cast < 1.9 then
				z.ID = aimedID
			end
		end
		
		local cost = s.SpellCost(z.ID)
		if a.Focus < s.SpellCost(s.SpellName(z.ID)) then
			return false
		end
		
		local _, regen = GetPowerRegen()
		local castTime = math.max(1, s.CastTime(s.SpellName(z.ID)))
		local focus = a.Focus - cost + castTime * regen
		local duration = a.CSCool - castTime
		local needed = c.GetCost("Chimera Shot")
		local addedTime = math.max(1, c.GetCastTime("Steady Shot"))
		local addedFocus = a.FocusAdded() + regen * addedTime
		while true do
			if focus >= needed then
				return true
			elseif duration <= 0 then
				return false
			end
			duration = duration - addedTime
			focus = focus + addedFocus
		end
	end
})

c.AddOptionalSpell("Rapid Fire", "if Talented", {
	CheckFirst = function()
		return c.HasTalent("Posthaste")
			and not c.HasBuff("Rapid Fire")
	end
})

c.AddOptionalSpell("Readiness", nil, {
	CheckFirst = function()
		return c.GetCooldown("Rapid Fire") > 60
	end
})

addSpell("Kill Command", "under Resistance is Futile", {
	NotIfActive = true,
	CheckFirst = function()
		return c.HasBuff("Resistance is Futile")
	end
})

local function shouldRefreshISS(shots)
	return shots * math.max(1, c.GetCastTime("Steady Shot")) > a.ISSDur
end

c.RegisterAura("Steady Shot", "Improved Steady Shot")

c.AddSpell("Steady Shot", "No Buff", {
	CheckFirst = function()
		return shouldRefreshISS(2)
	end
})

c.AddSpell("Steady Shot", "Refresh Buff", {
	CheckFirst = function()
		
		local castTime = math.max(1, c.GetCastTime("Steady Shot"))
--		if c.GetBuffDuration("Improved Steady Shot") > 2 * castTime + 2 then
--print("too soon!")
--			return false
--		end
		
		local _, regen = GetPowerRegen()
		local focusAdded = a.FocusAdded() + castTime * regen
		if a.NextSSIsImproved then
			castTime = 0
		else
			focusAdded = 2 * focusAdded
		end
		
		-- will it cap us?
		local wiggle = regen * math.min(c.GetCastTime("Aimed Shot"), 1.9)
		if a.Focus + focusAdded	> s.MaxPower("player") - wiggle then
			return false
		end
		
		-- will it interfere with Chimera Shot?
		if not underCarefulAim() and castTime > a.CSCool then
			return false
		end
		
		return true
	end
})

c.AddSpell("Steady Shot", "2", {
	CheckFirst = function()
		return a.NextSSIsImproved and shouldRefreshISS(1)
	end
})

c.AddSpell("Steady Shot", "2 AoE", {
	CheckFirst = function()
		if not a.NextSSIsImproved then
			return
		end
		
		local castTime = math.max(1, c.GetCastTime("Steady Shot"))
		local _, regen = GetPowerRegen()
		local focusAdded = a.FocusAdded() + castTime * regen
		return a.Focus + focusAdded < s.MaxPower("player") - regen
	end
})

---------------------------------------------------------------------- Survival
c.AddSpell("Explosive Shot", nil, {
	Override = function(z)
		return a.AllowES
			and c.GetCooldown("Explosive Shot") < .1
			and (a.Focus >= c.GetCost("Explosive Shot")
				or c.HasBuff("Lock and Load"))
	end
})

c.AddSpell("Explosive Shot", "for Free", {
	Override = function(z)
		return a.AllowES and c.HasBuff("Lock and Load")
	end
})

addOptionalSpell("Black Arrow")
